'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ClientWithPool = exports.Client = undefined;

var _net = require('net');

var _net2 = _interopRequireDefault(_net);

var _socketPool = require('socket-pool');

var _socketPool2 = _interopRequireDefault(_socketPool);

var _once = require('once');

var _once2 = _interopRequireDefault(_once);

var _decoder = require('./decoder');

var _decoder2 = _interopRequireDefault(_decoder);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Socket client with connection pool
const DEFAULT_BUFFER_LENGTH = 16;
const SOCKET_ERROR = 'SOCKET_ERROR';
const SOCKET_CLOSE_ERROR = 'SOCKET_CLOSE_ERROR';

const extraLength = chunk => {
  const arr = Array.prototype.slice.call(chunk.slice(0, 16));
  let i = 0;
  let extra = 0;

  while (i < 3) {
    extra += arr.pop() * Math.pow(256, i++);
  }

  return extra;
};

const throws = name => {
  throw new Error(`${name} must be implemented`);
};

class RequestBase {
  constructor({
    socket,
    host,
    port,
    buffer
  }) {

    this._socket = socket;
    this._host = host;
    this._port = port;
    this._buffer = buffer;
    this._resolve = this._reject = null;
    this._chunks = [];
    this._heap = null;
  }

  _write() {
    throws('_write');
  }

  _release() {
    throws('_release');
  }

  _decode() {
    (0, _decoder2.default)(this._heap).then(this._resolve, this._reject);
  }

  start() {
    return new Promise((resolve, reject) => {
      this._resolve = resolve = (0, _once2.default)(resolve);
      this._reject = reject = (0, _once2.default)(reject);

      let bufferLength = DEFAULT_BUFFER_LENGTH;
      const socket = this._socket;

      socket.on('error', err => {
        socket.destroy();
        err.code = SOCKET_ERROR;
        reject(err);
      });

      socket.on('data', chunk => {
        const chunks = this._chunks;

        if (!chunks.length) {
          bufferLength += extraLength(chunk);
        }

        chunks.push(chunk);

        const heap = this._heap = Buffer.concat(chunks);

        if (heap.length >= bufferLength) {
          this._done();
        }
      });

      this._write();
    });
  }
}

// Basic request for normal single-use socket connection
class Request extends RequestBase {
  _write() {
    const socket = this._socket;

    socket.on('close', err => {
      if (err) {
        err.code = SOCKET_CLOSE_ERROR;
        return this._reject(err);
      }

      this._decode();
    });

    socket.connect(this._port, this._host, () => {
      socket.write(this._buffer);
    });
  }

  _done() {
    this._socket.destroy();
  }
}

// Request for reusable socket
class RequestForPool extends RequestBase {
  _write() {
    this._socket.write(this._buffer);
  }

  _done() {
    this._socket.release();
    this._decode();
  }
}

class ClientBase {
  constructor(Request) {
    this._Request = Request;
  }

  _socket() {
    throws('_socket');
  }

  request(host, port, buffer) {
    return Promise.resolve(this._socket(host, port)).then(socket => new this._Request({
      socket,
      host,
      port,
      buffer
    }).start());
  }
}

// With normal single-use socket
class Client extends ClientBase {
  constructor() {
    super(Request);
  }

  _socket() {
    return new _net2.default.Socket();
  }
}

exports.Client = Client; // With socket pool

class ClientWithPool extends ClientBase {
  constructor(pool) {
    super(RequestForPool);

    this._pool = pool;
    this._pools = {};
  }

  _getPool(host, port) {
    const key = `${host}:${port}`;
    return this._pools[key] || (this._pools[key] = new _socketPool2.default({
      pool: this._pool,
      connect: {
        host,
        port
      }
    }));
  }

  // Acquire the socket connection from the pool
  _socket(host, port) {
    const pool = this._getPool(host, port);
    return pool.acquire();
  }
}
exports.ClientWithPool = ClientWithPool;